#include <string.h>
#include <errno.h>

#include <jni.h>
#include <bm_module.h>

#include "libcryptsetup.h"
#include "internal.h"

#include "cryptsetup-java.h"

#include <android/log.h>
#define APP_NAME "cryptsetup"

struct dump_usrptr
{
	char *str;
	int state;
	const char* const *filter;
};

#define TYPE_LUKS1 1
#define TYPE_LUKS2 2
#define TYPE_BITLK 3

#define DUMP_STR_SIZE (4*1024)

static void log_android(int level, const char *msg, void *usrptr)
{
	int alevel;
	if (level == CRYPT_LOG_VERBOSE)
		alevel = ANDROID_LOG_INFO;
	else if (level == CRYPT_LOG_ERROR)
		alevel = ANDROID_LOG_ERROR;
	else
		alevel = ANDROID_LOG_WARN;

	__android_log_write(alevel, APP_NAME, msg);
}

struct crypt_device* cryptsetup_open(block_t *block, int type, int *flags)
{
	struct and_device *dev = calloc(1, sizeof(struct and_device));
	if (!dev)
		return NULL;

	dev->header[0] = 'Z';
	dev->blk = block;
	dev->fd = -1;

	crypt_set_debug_level(CRYPT_LOG_DEBUG_JSON);

	struct crypt_device *cd = NULL;
	int r = crypt_init(&cd, (const char*)dev);
	if (r || !cd)
		goto error;

	crypt_set_userptr(cd, dev);
	crypt_set_log_callback(cd, log_android, NULL);

	const char *str_type;
	switch (type)
	{
	case TYPE_LUKS1: str_type = CRYPT_LUKS1; break;
	case TYPE_LUKS2: str_type = CRYPT_LUKS2; break;
	case TYPE_BITLK: str_type = CRYPT_BITLK; break;
	default: goto error;
	}

	r = crypt_load(cd, str_type, NULL);
	if (r)
	{
		log_err(cd, _("Device is not a valid %s device."), str_type);
		dev->flags |= FLAG_INCORRECT_FORMAT;
		goto error;
	}

	return cd;
	
error:
	if (flags)
		*flags = dev->flags;
	if (dev)
		free(dev);
	if (cd)
		crypt_free(cd);
	return NULL;
}

void* __UI cryptsetup_init(block_t *block, int type, const char *pwd, int *flags)
{
	struct crypt_device *cd = cryptsetup_open(block, type, flags);
	if (!cd)
		return NULL;

	struct and_device *dev = crypt_get_userptr(cd);

	int r = crypt_activate_by_passphrase(cd, "dev", CRYPT_ANY_SLOT, pwd, strlen(pwd), 0);
	if (r)
	{
		log_err(cd, _("Error on open device %d"), r);

		if (r == -EPERM)
			dev->flags |= FLAG_INCORRECT_PASSWORD;

		if (dev->bm_blk)
		{
			if (dev->bm_blk->fd == block)
				dev->bm_blk->fd = NULL;
			dev->bm_blk->freee(dev->bm_blk);
			dev->bm_blk = NULL;
		}

		if (flags)
			*flags = dev->flags;
	}

	crypt_free(cd);

	block_t *out = dev->bm_blk;
	free(dev);
	return out;
}

static inline int log_filter(const char *str, const char* const *filter, int *state)
{
	if (*state)
	{
		if (!strcmp("\n", str))
			*state = 0;
		return 0;
	}
	
	for (int i=0; filter[i]; i++)
	{
		int len = strlen(filter[i]);
		if (!strncmp(filter[i], str, len))
		{
			if (!strchr(str, '\n'))
				*state = 1;
			return 0;	
		}
	}
	
	return 1;
}

static void log_dump(int level, const char *msg, void *usrptr)
{
	log_android(level, msg, NULL);

	struct dump_usrptr *d = usrptr;
	if (usrptr && level == CRYPT_LOG_NORMAL && (!d->filter ||
					log_filter(msg, d->filter, &d->state)))
		strncat(d->str, msg, DUMP_STR_SIZE);
}

char* __UI cryptsetup_dump(block_t *block, int type)
{
	const char* const f_luks1[] = {"MK digest:", "MK salt:", "MK iterations:",
		"UUID:", "\tSalt:", "\tKey material offset:", "\tAF stripes:",
		"LUKS", 0 };
	const char* const f_luks2[] = {"UUID:", "\tSalt:", "\tArea offset",
		"\tArea length", "\tDigest:", 0 };	
	const char* const f_bitlk[] = {"\tSalt:", 0 };	
			
	struct crypt_device *cd = cryptsetup_open(block, type, NULL);
	if (!cd)
		return NULL;

	int r = -1;
	struct dump_usrptr usrptr;
	usrptr.str = malloc(DUMP_STR_SIZE);
	if (!usrptr.str)
		goto error;

	usrptr.state = 0;
	switch (type)
	{
	case TYPE_LUKS1: 
		usrptr.filter = f_luks1;
		strcpy(usrptr.str, "LUKS header information\n");
		break;
	case TYPE_LUKS2:
		usrptr.filter = f_luks2;
		usrptr.str[0] = '\0';
		break;
	case TYPE_BITLK:
		usrptr.filter = f_bitlk;
		usrptr.str[0] = '\0';
		break;
	default: goto error;
	}
	crypt_set_log_callback(cd, log_dump, &usrptr);

	r = crypt_dump(cd);
	if (r)
	{
		log_err(cd, _("Error on get dump %d"), r);
		goto error;
	} else
		usrptr.str[DUMP_STR_SIZE-1] = '\0';

error:
	if (r && usrptr.str)
	{
		free(usrptr.str);
		usrptr.str = NULL;
	}
	void *dev = crypt_get_userptr(cd);
	crypt_free(cd);
	if (dev) free(dev);
	return usrptr.str;
}

static int cryptsetup_bm_cloop(const char *cipher, const char *cipher_mode,
				 size_t volume_key_size,
				 double *encryption_mbs, double *decryption_mbs)
{
	int r, buffer_size = 1024 * 1024;

	do
	{
		r = crypt_benchmark(NULL, cipher, cipher_mode,
				    volume_key_size, 0, buffer_size,
				    encryption_mbs, decryption_mbs);
		if (r == -ERANGE)
		{
			if (buffer_size < 1024 * 1024 * 65)
				buffer_size *= 2;
			else
				r = -ENOENT;
		}
	} while (r == -ERANGE);

	return r;
}

static int benchmark_callback(uint32_t time_ms, void *usrptr)
{
	return 0;
}

#ifdef MBEDTLS_SELF_TEST
int mbedtls_sha1_self_test( int verbose );
int mbedtls_sha256_self_test( int verbose );
int mbedtls_sha512_self_test( int verbose );

int mbedtls_aes_self_test( int verbose );
int mbedtls_serpent_self_test( int verbose );
int mbedtls_twofish_self_test( int verbose );
#endif

int __UI cryptsetup_benchmark(const char *cipher, const char *hash, int key_size)
{
#ifdef MBEDTLS_SELF_TEST
	if (hash)
	{
		if (!strcmp("sha1", hash) && mbedtls_sha1_self_test(0))
			return -1;
		else if (!strcmp("sha256", hash) && mbedtls_sha256_self_test(0))
			return -1;
		else if (!strcmp("sha512", hash) && mbedtls_sha512_self_test(0))
			return -1;
	} else if (cipher)
	{
		if (!strcmp("aes-cbc", cipher) && mbedtls_aes_self_test(0))
			return -1;
		if (!strcmp("serpent-cbc", cipher) && mbedtls_serpent_self_test(0))
			return -1;
		if (!strcmp("twofish-cbc", cipher) && mbedtls_twofish_self_test(0))
			return -1;
	}
#endif

	if (cipher)
	{
		__android_log_print(ANDROID_LOG_WARN, APP_NAME, "benchmark %s", cipher);
		
		double enc_mbr = 0, dec_mbr = 0;
		char cipher_name[MAX_CIPHER_LEN], cipher_mode[MAX_CIPHER_LEN];

		int r = crypt_parse_name_and_mode(cipher, cipher_name, NULL, cipher_mode);
		if (r < 0)
		{
			__android_log_print(ANDROID_LOG_WARN, APP_NAME, "failed parse %s", cipher);
			return -1;
		}

		char *c;
		if ((c = strchr(cipher_mode, '-')))
			*c = '\0';

		__android_log_print(ANDROID_LOG_WARN, APP_NAME, "name '%s' mode '%s'", cipher_name, cipher_mode);

		r = cryptsetup_bm_cloop(cipher_name, cipher_mode, key_size, &enc_mbr, &dec_mbr);
		if (r)
		{
			__android_log_print(ANDROID_LOG_WARN, APP_NAME, "benchmark res %d", r);
			return -1;
		}

		return ((((int)(enc_mbr*8))&0xFFFF) << 16) | (((int)(dec_mbr*8))&0xFFFF);
	} else if (!hash)
	{
		return -1;
	} else if (strncmp(hash, "argon2", 6))
	{
		struct crypt_pbkdf_type pbkdf = {
			.type = CRYPT_KDF_PBKDF2,
			.hash = hash,
			.time_ms = 1000,
		};

		int r = crypt_benchmark_pbkdf(NULL, &pbkdf, "foo", 3, "bar", 3, key_size,
					&benchmark_callback, &pbkdf);
		if (r < 0)
			return -1;
		else
			return pbkdf.iterations;
	} else
	{
		struct crypt_pbkdf_type pbkdf = {
			.type = hash,
			.time_ms = DEFAULT_LUKS2_ITER_TIME,
			.max_memory_kb = DEFAULT_LUKS2_MEMORY_KB,
			.parallel_threads = DEFAULT_LUKS2_PARALLEL_THREADS,
		};

		int r = crypt_benchmark_pbkdf(NULL, &pbkdf, "foo", 3,
			"0123456789abcdef0123456789abcdef", 32,
			key_size, &benchmark_callback, &pbkdf);
		if (r < 0)
			return -1;
		else
			return pbkdf.iterations;
	}
}
